/*
 * imu-bno055.h
 *
 * Copyright (C) 2017 Kyle Crane
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef EMBEDEDROBOTLIBRARY_IMU_BNO055_H_
#define EMBEDEDROBOTLIBRARY_IMU_BNO055_H_

enum BNO055_ERR
{
  BNO_OK        = 0,
  ERR_BNO_GEN   = -11,
  ERR_BNO_NRDY  = -12,
  ERR_BNO_READ  = -13,
  ERR_BNO_WRITE = -20
};

typedef enum {BNO_RAW_ACC, BNO_RAW_GYRO, BNO_RAW_MAG} bno_raw_type;


/** @brief Representation of the calibration registers of the BNO055 */
struct BNO055_CalData
{
  int16_t acc_x_offset;
  int16_t acc_y_offset;
  int16_t acc_z_offset;
  int16_t mag_x_offset;
  int16_t mag_y_offset;
  int16_t mag_z_offset;
  int16_t gyr_x_offset;
  int16_t gyr_y_offset;
  int16_t gyr_z_offset;
  int16_t acc_radius;
  int16_t mag_radius;
};


/** @brief Orientation data in Euler representation.                */
struct EulerVector
{
  double heading;
  double pitch;
  double roll;
};


/** @brief Orientation dat in quaternion representation.             */
struct QuaternionVector
{
  double w;
  double x;
  double y;
  double z;
};


/** @brief Class to interface to the BNO055 IMU device.
 *
 *  This class encapsulates the BNO055 device registers and defines an
 *  interface to communication with it over an I2C bus.  Note that the
 *  device powers on in CONFIG mode which allows configuration registers
 *  to be changed.  Once switched to operation modes the device can not
 *  be further configured.  Use setDataSource() to determine if the
 *  object reads euler angles from the device directly or computes them
 *  from quaterions.
 *
 */
class BNO055 : public ISensor<EulerVector>, public I_I2C
{
public:
  enum SETTINGS
  {
    RETRY = 3,
    SOURCE_EULER        = false,
    SOURCE_QUAT         = true,
    TEMP_UNIT_C         = false,
    TEMP_UNIT_F         = true,
    CAL_SYS_STATUS      = 0xC0,
    CAL_GYR_STATUS      = 0x30,
    CAL_ACC_STATUS      = 0x0C,
    CAL_MAG_STATUS      = 0x03,
    SYS_RESET           = 0x20,
    PWR_MODE_NORMAL     = 0X00,
    PWR_MODE_LOWPOWER   = 0X01,
    PWR_MODE_SUSPEND    = 0X02,
    ST_ALL_PASSED       = 0x0F,

    OPR_MODE_CFG  = 0x00, OPR_MODE_ACC,     OPR_MODE_MAG,
    OPR_MODE_GYR,         OPR_MODE_ACCMAG,  OPR_MODE_ACCGYR,
    OPR_MODE_MAGGYR,      OPR_MODE_AMG,     OPR_MODE_IMU,
    OPR_MODE_COMPASS,     OPR_MODE_M4G,     OPR_MODE_NDOF_FMC_OFF,
    OPR_MODE_NDOF,

    TM_UNIT_BITS = 0x08, TM_UNIT_C = 0x00, TM_UNIT_F = 0x08,
  };

  enum REGMAP
  {
    CHIP_ID,          ACC_ID,           MAG_ID,           GYR_ID,
    SW_REV_ID_LSB,    SW_REV_ID_MSB,    BL_REV_ID,        PAGE_ID,
    ACC_X_LSB,        ACC_X_MSB,        ACC_Y_LSB,        ACC_Y_MSB,
    ACC_Z_LSB,        ACC_Z_MSB,        MAG_X_LSB,        MAG_X_MSB,
    MAG_Y_LSB,        MAG_Y_MSB,        MAG_Z_LSB,        MAG_Z_MSB,
    GYR_X_LSB,        GYR_X_MSB,        GYR_Y_LSB,        GYR_Y_MSB,
    GYR_Z_LSB,        GYR_Z_MSB,        EUL_HEADING_LSB,  EUL_HEADING_MSB,
    EUL_ROLL_LSB,     EUL_ROLL_MSB,     EUL_PITCH_LSB,    EUL_PITCH_MSB,
    QUA_W_LSB,        QUA_W_MSB,        QUA_X_LSB,        QUA_X_MSB,
    QUA_Y_LSB,        QUA_Y_MSB,        QUA_Z_LSB,        QUA_Z_MSB,
    LIA_X_LSB,        LIA_X_MSB,        LIA_Y_LSB,        LIA_Y_MSB,
    LIA_Z_LSB,        LIA_Z_MSB,        GRV_X_LSB,        GRV_X_MSB,
    GRV_Y_LSB,        GRV_Y_MSB,        GRV_Z_LSB,        GRV_Z_MSB,
    TEMP,             CALIB_STAT,       ST_RESULT,        INT_STA,
    SYS_CLK_STATUS,   SYS_STATUS,       SYS_ERR,          UNIT_SEL,
    RESERVED_3C,      OPR_MODE,         PWR_MODE,         SYS_TRIGGER,
    TEMP_SOURCE,      AXIS_MAP_CONFIG,  AXIS_MAP_SIGN,

    ACC_OFFSET_X_LSB=0x55,  ACC_OFFSET_X_MSB,     ACC_OFFSET_Y_LSB,
    ACC_OFFSET_Y_MSB,       ACC_OFFSET_Z_LSB,     ACC_OFFSET_Z_MSB,
    MAG_OFFSET_X_LSB,       MAG_OFFSET_X_MSB,     MAG_OFFSET_Y_LSB,
    MAG_OFFSET_Y_MSB,       MAG_OFFSET_Z_LSB,     MAG_OFFSET_Z_MSB,
    GYR_OFFSET_X_LSB,       GYR_OFFSET_X_MSB,     GYR_OFFSET_Y_LSB,
    GYR_OFFSET_Y_MSB,       GYR_OFFSET_Z_LSB,     GYR_OFFSET_Z_MSB,
    ACC_RADIUS_LSB,         ACC_RADIUS_MSB,       MAG_RADIUS_LSB,
    MAG_RADIUS_MSB
  };

public:
  uint    tx_errs;            /** Transmit error count.               */
  uint    rx_errs;            /** Receive error count.                */

protected:
  uint8_t m_address;          /** Device address.                     */
  bool    m_source;           /** Data source for orientation output  */
  int     m_lastError;        /** Last device error reported          */

  EulerVector       m_euler;  /** Contains Euler angle data representation.*/
  QuaternionVector  m_qt;     /** Contains Quaterion data representation.  */

public:
  BNO055(uint address=0x29);
  virtual ~BNO055(){};

  bool isReady();
  bool ping();
  bool isPresent();
  bool reset();
  void enable(bool ena=true);
  int  getLastError();
  bool write(uint16_t reg, uint8_t* byte, uint count);
  bool read(uint16_t reg, uint8_t* buffer, uint count);


  // Control interface methods
  void     setDataSource(bool source = false);
  bool     setOprMode(uint8_t mode);
  bool     setTempUnits(bool units);
  bool     setEulerUnits(bool units);
  bool     setGyroUnits(bool uints);
  bool     setAccelUnits(bool units);
  bool     getDataSource();
  uint8_t  getChipID();
  uint16_t getSoftwareRev();
  uint8_t  getBootRev();
  uint8_t  getOprMode();
  uint8_t  getSelfTest();
  uint8_t  getCalStatus();
  uint8_t  getSysError();
  uint8_t  getSysStatus();
  uint8_t  getUnits();

  bool    exportCalibration(BNO055_CalData& calibration);
  bool    importCalibration(BNO055_CalData& calibration);

  // Data interface methods
  virtual   void update();
  virtual   EulerVector get(unsigned force = 0);
  EulerVector           getEuler();
  QuaternionVector      getQuaternion();
  int8_t    getTemperature();
  double    getHeading();
  double    getPitch();
  double    getRoll();
  SenTriInt getRaw(bno_raw_type sel);

public:
  bool     writeByte(uint8_t reg, uint8_t val);
  bool     writeWord(uint8_t reg, uint16_t val);
  uint8_t  readByte(uint8_t reg);
  uint16_t readWord(uint8_t reg);
};


/** @brief Create the IMU sensor device using supplied parameters
 *
 *  @param I_I2C& pBus: Reference to I2C bus object following I_I2C interface.
 *  @param uint address: The address of the IMU on the I2C bus.
 */
inline BNO055::BNO055(uint address)
{
  if (address == 0x28 or address == 0x29)
    m_address = address;
  else
    m_address = 0;

  m_source       = false;
  m_enable       = true;
  m_lastError    = 0;
  tx_errs        = 0;
  rx_errs        = 0;
}


/** @brief write byte sized register on device.
 *
 *  @param uint8_t reg: Register address.
 *  @param uint8_t val: The value to store in the register.
 *  @return int: 0 on success, error otherwise.
 */
inline bool BNO055::write(uint16_t reg, uint8_t* byte, uint count)
{
  int     err     = 0;
  uint    sent    = 0;
  uint8_t lsb     = reg;

  m_lastError  = BNO_OK;

  if (!isReady())
  {
    m_lastError = ERR_BNO_NRDY;
    tx_errs++;
    return true;
  }

  Wire.beginTransmission(m_address);
  sent += Wire.write(lsb);
  sent += Wire.write(byte, count);
  err = Wire.endTransmission();

  // Errors:
  // -21 = Data too long
  // -22 = NAK on address
  // -23 = NAK on data
  // -24 = Other write
  // -25 = Write count mis-match
  if(err)
  {
    m_lastError = ERR_BNO_WRITE - err;
    tx_errs++;
    return true;
  }
  else if(sent != (count+1))
  {
    m_lastError = ERR_BNO_WRITE - 5;
    tx_errs++;
    return true;
  }

  return false;
}


/** @brief Write word sized register on device.
 *
 *  @param uint8_t reg: Register address.
 *  @param uint16_t val: The value to store in the register.
 *  @return int: 0 on success, error otherwise.
 */
inline bool BNO055::writeByte(uint8_t reg, uint8_t val)
{
  int err   = 0;

  err = write(reg, &val ,1);

  return err;
}


/** @brief Write word sized register on device.
 *
 *  @param uint8_t reg: Register address.
 *  @param uint16_t val: The value to store in the register.
 *  @return int: 0 on success, error otherwise.
 */
inline bool BNO055::writeWord(uint8_t reg, uint16_t val)
{
  int err   = 0;

  err = write(reg, (uint8_t*)&val ,2);

  return err;
}



/** @brief Read data bytes from the I2C bus.
 *
 *  @param uint8_t reg: The register address to read from.
 *  @param uint8_t* buffer: Buffer to read the bytes into.
 *  @param uint count: The number of bytes to be read.
 *  @return bool: 0=success, 1 failure.
 */
inline bool BNO055::read(uint16_t reg, uint8_t* buffer, uint count)
{
  int     err     = 0;
  uint    rx_read = 0;
  uint8_t lsb     = reg;

  m_lastError = 0;

  if (!isReady())
  {
    m_lastError = ERR_BNO_NRDY;
    rx_errs++;
    return true;
  }

  Wire.beginTransmission(m_address);
  Wire.write(lsb);
  err = Wire.endTransmission(false);

  rx_read = Wire.requestFrom(m_address, count, true);
  for (uint8_t i=0; i<count; i++)
  {
    buffer[i] = Wire.read();
  }

  // Errors:
  // -14 = Data too long
  // -15 = NAK on address
  // -16 = NAK on data
  // -17 = Other write
  // -18 = Read count mis-match
  if (err)
  {
    m_lastError = ERR_BNO_READ - err;
    rx_errs++;
    return true;
  }
  if (rx_read != count)
  {
    m_lastError = ERR_BNO_READ - 5;
    rx_errs++;
    return true;
  }

  return false;
}


/** @brief Read byte sized register on device.
 *
 *  @param uint8_t reg: Register address.
 *  @return uint8_t: Register value.
 */
inline uint8_t BNO055::readByte(uint8_t reg)
{
  uint8_t value = 0;

  read(reg, &value, 1);

  return value;
}


/** @brief Read word sized register on device.
 *
 *  @param uint8_t reg: Register address.
 *  @return uint16_t: Register value.
 */
inline uint16_t BNO055::readWord(uint8_t reg)
{
  uint16_t  value = 0;

  read(reg, (uint8_t*)&value, 2);

  return value;
}







/** @brief Determine if the device is properly configured and ready to use.
 *
 *  @return bool: True if ready, false otherwise.
 */
inline bool BNO055::isReady()
{
  if (m_enable and m_address > 0)
    return true;
  else
    return false;
}


/** @brief Query the bus for the device.
 *
 *  @return bool: True if address is ACK'ed, false otherwise.
 */
inline bool BNO055::ping()
{
  bool ping = false;

  Wire.beginTransmission(m_address);
  if (Wire.endTransmission() == 0)
    ping = true;

  return ping;
}


/** @brief Query the bus for the device.
 *
 *  @return bool: True if chip_id is found, false otherwise.
 */
inline bool BNO055::isPresent()
{
  if (getChipID() == 0xA0)
    return true;
  else
    return false;
}


/** @brief Set the system reset bit and cause the unit to reset.
 *
 *  Note that the device may not respond for 650ms or more after a reset.
 *  ping() polling or isPresent() polling can be used to determine when
 *  the device comes back online.
 *
 *  @return bool: 0 on sucess, 1 failure.
 */
inline bool BNO055::reset()
{
  bool err = writeByte(SYS_TRIGGER, 0x20);
  return err;
}



/** @brief Set or clear the enalbe flag for this object.
 *
 *  @param bool ena: 1=Enable | 0=Disable.
 */
inline void BNO055::enable(bool ena)
{
  m_enable = ena;

  if (!m_enable)
  {
    m_qt.w = m_qt.x = m_qt.y = m_qt.z = 0;
    m_euler.heading = m_euler.pitch = m_euler.roll = 0;
  }
}


/** @brief Returns the value of the last error indicator from the I2C bus.
 *
 *  @return int: Zero no error, any other value signals error on last operation.
 */
inline int BNO055::getLastError()
{
  return m_lastError;
}










/** @brief Set the type of data used to represent position.
 *
 *  @param bool source: 1=quaternion | 0=euler (default)
 */
inline void BNO055::setDataSource(bool source)
{
  m_source = source;
}


/** @brief Initialize or re-initialize the device and operation mode.
 *
 *  @param uint8_t mode = Device operation mode to switch to.
 *  @return bool: 0 on success without error, 1 on any error.
 */
inline bool BNO055::setOprMode(uint8_t mode)
{
  int result;

  result = writeByte(OPR_MODE, mode);
  delay (50);

  if (result)
    return true;
  else
    return false;
}


/** @brief Set the temperature units for temperature reporting
 *
 *  @param bool units: 0 for C, 1 for F
 *  @return bool: false = success, true = error
 */
inline bool BNO055::setTempUnits(bool units)
{
  uint8_t regVal = getUnits();
  uint8_t mask   = 0;
  uint8_t newVal = 0;
  int     err    = 0;

  // If we get a bad value then eject.
  if (regVal > 0x97)
    return true;

  if (units)
    newVal = regVal | 0x10;  // 0001.0000 bit4
  else
    newVal = regVal & 0xEF;  // 1110.1111 bit4

  // Write the new register value and report any error.
  err = writeByte(UNIT_SEL, newVal);
  if (err)
    return true;

  return false;
}


/** Sets the units for euler data from the IMU.
 *
 *  @param bool units: false=degrees, true=radians.
 */
inline bool BNO055::setEulerUnits(bool units)
{
  uint8_t regVal = getUnits();
  uint8_t mask   = 0;
  uint8_t newVal = 0;
  int     err    = 0;

  // If we get a bad value then eject.
  if (regVal > 0x97)
    return true;

  if (units)
    newVal = regVal | 0x04; // 0000.1000 bit3
  else
    newVal = regVal & 0xFB; // 1111.0111 bit3

  // Write the new register value and report any error.
  err = writeByte(UNIT_SEL, newVal);
  if (err)
    return true;

  return false;
}


inline bool BNO055::getDataSource()
{
  return m_source;
}


/** @brief Obtain the value of the CHIP_ID register.
 *
 *  @return uint8_t: Return the CHIP_ID register value.
 */
inline uint8_t BNO055::getChipID()
{
  uint8_t result = readByte(CHIP_ID);

  return result;
}


/** @brief Obtain the value of the SW_REV_ID registers.
 *
 *  @return uint16_t: The SW_REV_ID register word value.
 */
inline uint16_t BNO055::getSoftwareRev()
{
  uint16_t result  = readWord(SW_REV_ID_LSB);

  return result;
}


/** @brief Obtain the value of the BL_REV registers.
 *
 *  @return uint8_t: The BL_REV register word value.
 */
inline uint8_t BNO055::getBootRev()
{
  uint8_t result = readByte(BL_REV_ID);

  return result;
}


/** @brief Returns the current operation mode of the device.
 *
 *  @return uint8_t: The register value.
 */
inline uint8_t BNO055::getOprMode()
{
  uint8_t result = readByte(OPR_MODE);

  return result;
}


/** @brief Returns the self-test results from the device.
 *
 *  @return uint8_t: The register value.
 */
inline uint8_t BNO055::getSelfTest()
{
  uint8_t result = readByte(ST_RESULT);

  return result;
}


/** @brief Returns the calibration status of the 3 sensors and device.
 *
 *  @return uint8_t: The register value.
 */
inline uint8_t BNO055::getCalStatus()
{
  uint8_t result = readByte(CALIB_STAT);

  return result;
}


/** @brief Returns the value stored in the system error register.
 *
 *  @return uint8_t: The register value.
 */
inline uint8_t BNO055::getSysError()
{
  uint8_t result = readByte(SYS_ERR);

  return result;
}


/** @brief Returns the value stored in the system error register.
 *
 *  @return uint8_t: The register value.
 */
inline uint8_t BNO055::getSysStatus()
{
  uint8_t result = readByte(SYS_STATUS);

  return result;
}


/** @brief Returns the value stored in the unit selection register.
 *
 *  @return uint8_t: The register value.
 */
inline uint8_t BNO055::getUnits()
{
  uint8_t result = readByte(UNIT_SEL);

  return result;
}


/** @brief Update object internal data from the device over I2C bus.
 *
 *  This method triggers an update from the sensor of all needed data
 *  to represent orientation.  If the object is set to use Euler values
 *  as the source then only the Euler registers are read.  If the object
 *  is configured to use quaterion data then only those registers are
 *  read and Euler data is calculated from them.
 */
inline void BNO055::update()
{
  int err = 0;

  // Read the appropriate data from the device depending on m_source. If
  // m_source is set then read quaternion registers.  If not then read
  // euler angle registers.

  if (m_source)
  {
    int16_t wReg  = 0;
    int16_t xReg  = 0;
    int16_t yReg  = 0;
    int16_t zReg  = 0;
    uint8_t data[8];
    memset(data, 0, 8);

    err = read(QUA_W_LSB, data, 8);

    if (err)
    {
      m_qt.w = m_qt.x = m_qt.y = m_qt.z = 0;
      m_euler.heading = m_euler.pitch = m_euler.roll = 0;
      return;
    }

    // Place the bytes LSB | MSB
    wReg = (data[0] | (data[1] << 8));
    xReg = (data[2] | (data[3] << 8));
    yReg = (data[4] | (data[5] << 8));
    zReg = (data[6] | (data[7] << 8));

    // Convert to proper range and store in result.
    const double scale = (1.0 / 8192.0);

    m_qt.w = (double)(wReg/16384.0);
    m_qt.x = (double)(xReg/16384.0);
    m_qt.y = (double)(yReg/16384.0);
    m_qt.z = (double)(zReg/16384.0);
  }
  else
  {
    int16_t headingReg  = 0;
    int16_t rollReg     = 0;
    int16_t pitchReg    = 0;
    uint8_t data[6];
    memset(data, 0, 6);

    err = read(EUL_HEADING_LSB, data, 6);

    if (err)
    {
      m_qt.w = m_qt.x = m_qt.y = m_qt.z = 0;
      m_euler.heading = m_euler.pitch = m_euler.roll = 0;
      return;
    }

    // Place the bytes LSB | MSB
    headingReg = data[0] | (data[1] << 8);
    rollReg    = data[2] | (data[3] << 8);
    pitchReg   = data[4] | (data[5] << 8);

    // Convert to proper range.
    m_euler.heading = headingReg/16.0;
    m_euler.roll    = rollReg/16.0;
    m_euler.pitch   = pitchReg/16.0;
  }

  // If source is set then convert quaternion vector to euler angels and
  // update the euler vector internal object as well.

  if (m_source)
  {
    double sqw = m_qt.w*m_qt.w;
    double sqx = m_qt.x*m_qt.x;
    double sqy = m_qt.y*m_qt.y;
    double sqz = m_qt.z*m_qt.z;


    // Convert to euler notation and store in result.
    m_euler.heading  = atan2(2.0*(m_qt.x*m_qt.y+m_qt.z*m_qt.w),( sqx-sqy-sqz+sqw));
    m_euler.pitch    = asin(-2.0*(m_qt.x*m_qt.z-m_qt.y*m_qt.w)/( sqx+sqy+sqz+sqw));
    m_euler.roll     = atan2(2.0*(m_qt.y*m_qt.z+m_qt.x*m_qt.w),(-sqx-sqy+sqz+sqw));
  }
}


/** @brief Generic get function always returns the current euler vector.
 *
 *  @result EulerVector: A copy of the internal euler vector.
 */
inline EulerVector BNO055::get(unsigned force)
{
  if (force)
  {
    update();
  }

  return m_euler;
}


/** @brief Gets Euler angles from the IMU object.
 *
 *  @result EulerVector: A copy of the internal euler vector.
 */
inline EulerVector BNO055::getEuler()
{
  return m_euler;
}


/** @brief Returns the quaternion values from the IMU object.
 *
 *  @result QuaternionVector:  A copy of the internal quat vector.                   .
 */
inline QuaternionVector BNO055::getQuaternion()
{
  return m_qt;
}


/** @brief Obtain the temperature reading from the IMU.
 *
 *  @return int8_t: The temperature data in units defined by units config reg.
 */
inline int8_t BNO055::getTemperature()
{
  uint8_t result = readByte(TEMP);

  return result;
}


/** @brief Obtain the heading reading from the IMU.
 *
 *  @return int16_t: The heading data in degrees or radians see units reg.
 */
inline double BNO055::getHeading()
{
  return m_euler.heading;
}


/** @brief Obtain the pitch reading from the IMU.
 *
 *  @return int16_t: The pitch data in degrees or radians see units reg.
 */
inline double BNO055::getPitch()
{
  return m_euler.pitch;
}


/** @brief Obtain the roll reading from the IMU.
 *
 *  @return int16_t: The roll data in degrees or radians see units reg.
 */
inline double BNO055::getRoll()
{
  return m_euler.roll;
}


/** @brief Returns the raw sensor readings from any of the 3 sensors.
 *
 *  @param bno_raw_type sel: The sensor selected to read from.
 *  @return SenTriInt: Three integer values for x/y/z of the selected sensor.
 */
inline SenTriInt BNO055::getRaw(bno_raw_type sel)
{
  SenTriInt result;
  uint8_t data[6];
  bool    err = false;
  memset(data, 0, 6);

  switch (sel)
  {
    case BNO_RAW_ACC:
      err = read(ACC_X_LSB, data, 6);
      break;
    case BNO_RAW_MAG:
      err = read(MAG_X_LSB, data, 6);
      break;
    case BNO_RAW_GYRO:
      err = read(GYR_X_LSB, data, 6);
      break;
  }

  if (err)
  {
    result.x = 0;
    result.y = 0;
    result.z = 0;
  }
  else
  {
    result.x = (int16_t) (data[0] | (data[1] << 8));
    result.y = (int16_t) (data[2] | (data[3] << 8));
    result.z = (int16_t) (data[4] | (data[5] << 8));
  }

  return result;
}


/** @brief Extract the calibration configuration from the device.
 *
 *  @param BNO055_CalData& calibration: Calibration values to read.
 *  @return bool: 0=Success, all else failure.
 */
inline bool BNO055::exportCalibration(BNO055_CalData& calibration)
{
  int err = 0;
  uint8_t data[23];
  memset(data, 0, 22);

  err = read(ACC_OFFSET_X_LSB, data, 22);

  if (err)
    return true;

  // Assemble the data into the structure fields.
  calibration.acc_x_offset  = data[0]  | (data[1]  << 8);
  calibration.acc_y_offset  = data[2]  | (data[3]  << 8);
  calibration.acc_z_offset  = data[4]  | (data[5]  << 8);
  calibration.mag_x_offset  = data[6]  | (data[7]  << 8);
  calibration.mag_y_offset  = data[8]  | (data[9]  << 8);
  calibration.mag_z_offset  = data[10] | (data[11] << 8);
  calibration.gyr_x_offset  = data[12] | (data[13] << 8);
  calibration.gyr_y_offset  = data[14] | (data[15] << 8);
  calibration.gyr_z_offset  = data[16] | (data[17] << 8);
  calibration.acc_radius    = data[18] | (data[19] << 8);
  calibration.mag_radius    = data[20] | (data[21] << 8);

  return false;
}


/** @brief Install the calibration configuration onto the device.
 *
 *  @param BNO055_CalData& calibration: Calibration values to write.
 *  @return bool: 0=Success, all else failure.
 */
inline bool BNO055::importCalibration(BNO055_CalData& calibration)
{
  int err = 0;
  uint8_t data[23];
  memset(data, 0, 22);

  // load the data array with the provided calibration data.
  data[0]  =  calibration.acc_x_offset;
  data[1]  = (calibration.acc_x_offset >> 8);
  data[2]  =  calibration.acc_y_offset;
  data[3]  = (calibration.acc_y_offset >> 8);
  data[4]  =  calibration.acc_z_offset;
  data[5]  = (calibration.acc_z_offset >> 8);

  data[6]  =  calibration.mag_x_offset;
  data[7]  = (calibration.mag_x_offset >> 8);
  data[8]  =  calibration.mag_y_offset;
  data[9]  = (calibration.mag_y_offset >> 8);
  data[10] =  calibration.mag_z_offset;
  data[11] = (calibration.mag_z_offset >> 8);

  data[12] =  calibration.gyr_x_offset;
  data[13] = (calibration.gyr_x_offset >> 8);
  data[14] =  calibration.gyr_y_offset;
  data[15] = (calibration.gyr_y_offset >> 8);
  data[16] =  calibration.gyr_z_offset;
  data[17] = (calibration.gyr_z_offset >> 8);

  data[18] =  calibration.acc_radius;
  data[19] = (calibration.acc_radius >> 8);
  data[20] =  calibration.mag_radius;
  data[21] = (calibration.mag_radius >> 8);


  err = write(ACC_OFFSET_X_LSB, data, 22);

  if (err)
    return true;

  return false;
}

#endif /* EMBEDEDROBOTLIBRARY_IMU_BNO055_H_ */
